/* jshint undef: true, esnext: true */
/* global plugin, require, Packages, com, java, javax, org */

var _ = require("lodash");
var on = require("on")(plugin);
var Objects = require("convenience").Objects;
var getStackTraceAsString = require("convenience").Errors.getStackTraceAsString;
var facade = require("facade");
var OSSystemState = Packages.cattura.utils.os.OSSystemState;
var StringUtilities = Packages.cattura.utils.string.StringUtilities;
var EventType = Packages.cattura.agents.configuration.model.vos.events.EventType;

var EmailBuilder = org.simplejavamail.email.EmailBuilder;
var TransportStrategy = org.simplejavamail.api.mailer.config.TransportStrategy;
var MailerBuilder = org.simplejavamail.mailer.MailerBuilder;

var FileDataSource = javax.activation.FileDataSource;
var STORE = plugin.info().dataStore;
var settings = plugin.info().settings;
var LOGGER = plugin.info().logger;
var StringWriter = java.io.StringWriter;
var createAttachmentForIncident = require("./createAttachmentForIncident");

var Level = java.util.logging.Level;

var ThreadFactoryBuilder = com.google.common.util.concurrent.ThreadFactoryBuilder;
var addCallback = com.google.common.util.concurrent.Futures.addCallback;
var FutureCallback = com.google.common.util.concurrent.FutureCallback;
var successfulAsList = com.google.common.util.concurrent.Futures.successfulAsList;
var listeningDecorator = com.google.common.util.concurrent.MoreExecutors.listeningDecorator;
var newCachedThreadPool = java.util.concurrent.Executors.newCachedThreadPool;

var Optional = java.util.Optional;
var EventHelper = Packages.cattura.agents.configuration.model.utilities.EventHelper;

/*
 * Executor used for sending emails
 */
var mailExecutor = listeningDecorator(newCachedThreadPool(new ThreadFactoryBuilder()
    .setNameFormat("Email Notifier Plugin executor thread %d")
    .setDaemon(true)
.build()));

/*
 * Email notifiers
 */
var notifiers = _(Objects.toJS(STORE.select("notifier:*")))
    .map(notifier => [notifier.id, new Notifier(notifier)])
    .fromPairs()
    .value();

/*
 * This plugin will require an SMTP server to talk to.
 * By default it will use localhost; thus, assuming that
 * an SMTP server is running on the recorder. The standard
 * distribution comes with OpenSMTPd.
 */
var SSL_PORT     = 465;
var TLS_PORT     = 587;
var DEFAULT_PORT = 25;

function makeMailer ()
{
    var host = getStringSetting("host").orElse("localhost");
    var builder = MailerBuilder.withSMTPServerHost(host);
    
    var port = +getStringSetting("port").orElse(0);
    if (port)
    {
        builder.withSMTPServerPort(port);
    }
    
    var ssl = Objects.toJS(getBooleanSetting("ssl").orElse(port === SSL_PORT));
    var tls = Objects.toJS(getBooleanSetting("tls").orElse(port === TLS_PORT));
    var transportStrategy = ssl ? TransportStrategy.SMTPS
                                : tls ? TransportStrategy.SMTP_TLS
                                      : TransportStrategy.SMTP;
    
    if (transportStrategy)
    {
        builder.withTransportStrategy(transportStrategy);
    }
    
    var username = getStringSetting("username").orElse(null);
    if (username)
    {
        builder.withSMTPServerUsername(username);
    }
    
    var password = getStringSetting("password").orElse(null);
    if (password)
    {
        builder.withSMTPServerPassword(password);
    }
    
    builder.withDebugLogging(getBooleanSetting("debug").orElse(false));
    return builder.buildMailer();
}

/*
 * By default this plugin will handle anything that's not an internal event,
 * or is on the allowlist, or is not on the blocklist.
 */
var EVENT_TYPE_ALLOWLIST = [];
var EVENT_TYPE_BLOCKLIST = ["AudioLevelsEvent"];

function eventIsNotifiable (event)
{
    var helper = new EventHelper(event);
    return (
        !!_(EVENT_TYPE_ALLOWLIST).find(eventType => helper.isOfType(eventType)) ||
        !(
            helper.isInternalEvent() || 
            !!_(EVENT_TYPE_BLOCKLIST).find(eventType => helper.eventIsOfType(eventType))
        )
    );
}

function Notifier (data)
{
    _(this).extend(data).value();

    this.canHandleEvent = function canHandleEvent (event)
    {
        if (!eventIsNotifiable(event))
        {
            return false;
        }

        return _(this.events).find(function (type) {
            return type == event.eventType.name() || type == event.getClass().simpleName;
        });
    };

    this.handleEvent = function handleEvent (event, attachments)
    { 
        var t = this;
        var replyToAddress = getStringSetting("replyToAddress").orElse("cattura@" + OSSystemState.getMachineName());
        var replyToName = getStringSetting("replyToName").orElse("Cattura");
        var fromAddress = getStringSetting("fromAddress").orElse("cattura@" + OSSystemState.getMachineName());
        var fromName = getStringSetting("fromName").orElse("Cattura");

        return mailExecutor['submit(java.lang.Runnable)'](function () {
            var builder = EmailBuilder.startingBlank()
                .from(fromName, fromAddress)
                .to(t.email, t.email)
                .withSubject("[" + getUnitName() + "] " + event.description)
                .withPlainText(event.extendedDescription || event.description)
                .withReplyTo(replyToName, replyToAddress);
            
            if (event.eventType.name() == "EXCEPTION")
            {
                if (event.incident)
                {
                    attachExceptionText(event, builder);
                }
                else
                {
                    return; // don't want to handle this if it's a legacy ExceptionEvent
                }
            }
            

            addAttachments(builder, attachments);
            makeMailer().sendMail(builder.buildEmail());
        });
    };

    function addAttachments (builder, attachments)
    {
        if (!attachments)
        {
            return;
        }

        for (var filename in attachments)
        {
            try
            {
                builder.withAttachment(filename, new FileDataSource(attachments[filename]));
            }
            catch (e)
            {
                LOGGER.warning("Failed to attach " + filename + " to event report");
            }
        }
    }

    function attachExceptionText (event, builder)
    {
        if (event.incident.exception)
        {
            builder.withPlainText(getStackTraceAsString(event.incident.exception));
        }
    }
}

function testNotifier (email, events, event)
{
    new Notifier({
        email: email,
        events: events
    }).handleEvent(event);
}

function addNotifier (id, email, events)
{
    var notifier = {
        id: id,
        email: email,
        events: events
    };
    
    STORE.set("notifier:" + id, notifier);
    
    notifier = new Notifier(notifier);
    notifiers[id] = notifier;
    return notifier;
}

function removeNotifier (id)
{
    delete notifiers[id];
    STORE.remove("notifier:" + id);
}

function getNotifiersForEvent (event)
{
    if (!eventIsNotifiable(event))
    {
        return [];
    }
    
    return _(notifiers)
        .filter(notifier => notifier.canHandleEvent(event))
        .value();
}

function generateIdForNotifier ()
{
    var id = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
    if (_(notifiers).find(notifier => notifier.id == id))
    {
        return generateIdForNotifier();
    }

    return id;
}

// Find first valid Optional, or return empty Optional if none are valid
function firstPresent ()
{
    return _(arguments).find(argument => argument.isPresent()) || Optional.empty();
}

// Get user setting, then fall back to system setting
function getStringSetting (name)
{
    return firstPresent(
        plugin.userSettings().getString(name), plugin.systemSettings().getString(name)
    );
}

function getBooleanSetting (name)
{
    return firstPresent(
        plugin.userSettings().getBoolean(name), plugin.systemSettings().getBoolean(name)
    );
}

function getUnitName ()
{
    var settings = facade.retrieveProxy("SettingsProxy");
    var unitName = settings.getUserSettings().getString("unit-name").orElse("");
    if (!unitName.length())
    {
        unitName = OSSystemState.getMachineName();
    }
    return unitName;
}

function getExceptionFromEvent (event)
{
    if ('incident' in event)
    {
        return event.incident.exception;
    }
    else if ('exception' in event)
    {
        return event.exception;
    }
}

function generateAttachmentsForEvent (event)
{
    var attachments = {};
    if (!(event.eventType.name() == "EXCEPTION" && event.incident))
    {
        return attachments;
    }

    try
    {
        var attachment = createAttachmentForIncident(event.incident);
        if (attachment)
        {
            attachments["catturalogs.zip"] = attachment;
        }
    }
    catch (e)
    {
        LOGGER.warning("Failed to generate attachment for error report: " + getStackTraceAsString(e));
    }

    return attachments;
}

function cleanupAttachments (attachments)
{
    if (!attachments)
    {
        return;
    }

    _(attachments).each(function (attachment) {
        attachment.delete();
    });
}

plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "email", description: "Email address", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"}
    ],
    handler: function (email, events) {
        return addNotifier(generateIdForNotifier(), email, events);
    }
});

plugin.service({
    name: "addNotifier",
    description: "Add a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"},
        {name: "email", description: "Email address", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"}
    ],
    handler: addNotifier
});

plugin.service({
    name: "testNotifier",
    description: "Test a notifier with a test event.",
    parameters: [
        {name: "email", description: "Email address", type: "java.lang.String"},
        {name: "events", description: "Events to notify on", type: "java.util.List"},
        {name: "event", description: "Test event", type: "java.lang.Object"}
    ],
    handler: testNotifier
});

plugin.service({
    name: "removeNotifier",
    description: "Remove a notifier",
    parameters: [
        {name: "id", description: "Unique identifier for notifier", type: "java.lang.String"}
    ],
    handler: removeNotifier
});

plugin.service({
    name: "getNotifiers",
    description: "Return all registered notifiers",
    parameters: [],
    handler: () => _(notifiers).values().value()
});

function handleExceptionFromHandler (event, exception)
{
    var unwrappedException = exception.javaException || exception.rhinoException || exception;
    if (getExceptionFromEvent(event))
    {
        LOGGER.log(Level.WARNING, "Email Notification Plugin encountered (and swallowed) exception while trying to handle exception event", unwrappedException);
        return;
    }

    facade.sendNotification("HANDLE_EXCEPTION", new java.lang.RuntimeException("Email Notification Plugin encountered exception while trying to handle event: " + event, unwrappedException));
}

plugin.onEvent(function(event) {
    try
    {
        var notifiers = getNotifiersForEvent(event);
        if (!notifiers.length)
        {
            return;
        }

        /*
         * Create attachments and transmit events using executor.
         * Emails are sent in parallel, then when all emails are sent
         * the attachments are cleaned up.
         */
        mailExecutor.execute(function () {
            try
            {
                var attachments = generateAttachmentsForEvent(event);
                addCallback(successfulAsList(_(notifiers).map(function (notifier) {
                    var emailSent = notifier.handleEvent(event, attachments);
                    addCallback(emailSent, new FutureCallback({
                        onSuccess ()
                        {
                            LOGGER.log(Level.INFO, "Sent email notification of " + event.eventType + " event to " + notifier.email);
                        },
                        onFailure (exception)
                        {
                            handleExceptionFromHandler(event, exception);
                        }
                    }), mailExecutor);
                    return emailSent;
                }).value()), new FutureCallback({
                    onSuccess ()
                    {
                        //cleanupAttachments(attachments);
                    },
                    onFailure (exception)
                    {
                        //cleanupAttachments(attachments);
                        handleExceptionFromHandler(event, exception);
                    }
                }), mailExecutor);
            }
            catch (exception)
            {
                handleExceptionFromHandler(event, exception);
            }
        });
    }
    catch (exception)
    {
        handleExceptionFromHandler(event, exception);
    }
});